"""Read data from files."""

import json
import logging
from zipfile import ZipFile

import numpy as np
import pandas as pd

from pygt3x import Types
from pygt3x.activity_payload import (
    read_activity1_payload,
    read_activity2_payload,
    read_activity3_payload,
    read_nhanse_payload,
    read_temperature_payload,
)
from pygt3x.components import Header, Info, RawEvent


class FileReader:
    """Read GT3X/AGDC files.

    Parameters:
    -----------
    file_name:
        Input file name
    """

    def __init__(self, file_name: str):
        """Initialise."""
        self.file_name = file_name
        self.acceleration = np.empty((0, 4))
        self.temperature = np.empty((0, 3))

    def __enter__(self):
        """Open zipped file and ret up readers."""
        self.zipfile = ZipFile(self.file_name)
        try:
            self.logfile = self.zipfile.open("log.bin", "r")
            self.logreader = LogReader(self.logfile)
        except KeyError:
            # V1 file
            self.logreader = None
            self.logfile = self.zipfile.open("log.txt", "r")
            self.activity_file = self.zipfile.open("activity.bin", "r")
        self.info = Info(self.zipfile)
        self.calibration = self.read_json("calibration.json")
        self.temperature_calibration = self.read_json("temperature_calibration.json")

        return self

    def __exit__(self, typ, value, traceback):
        """Close file descriptors."""
        self.logfile.__exit__(typ, value, traceback)
        self.zipfile.__exit__(typ, value, traceback)

    def read_json(self, file_name):
        """Read calibration info from file."""
        if file_name not in self.zipfile.namelist():
            return None
        with self.zipfile.open(file_name) as f:
            calibration = json.load(f)
            return calibration

    def _fill_ism(self, idle_sleep_mode_started, idle_sleep_mode_ended, last_values):
        """Fill in gaps created by idle sleep mode."""
        timestamps = (
            np.arange(idle_sleep_mode_started, idle_sleep_mode_ended)
            .repeat(self.info.sample_rate)
            .reshape(-1, 1)
        )
        values = last_values.reshape((1, 3)).repeat(timestamps.shape[0], axis=0)

        return np.concatenate((timestamps, values), axis=1)

    def read_events(self, num_rows=None):
        """Read events from file.

        Parameters:
        -----------
        num_rows
            Number of events to read.
        """
        if num_rows is None:
            raw_event = self.logreader.read_event()
            while raw_event is not None:
                yield raw_event
                raw_event = self.logreader.read_event()
        else:
            for _ in range(0, num_rows):
                raw_event = self.logreader.read_event()
                yield raw_event

    def get_data(self, num_rows=None):
        """Yield acceleration data.

        Parameters:
        -----------
        num_rows
            Number of events to read.
        """
        acceleration = []
        temperature = []
        if self.logreader:
            idle_sleep_mode_started = None
            # This is used for filling in gaps created by idle sleep mode
            last_values = None
            for evt in self.read_events(num_rows):
                try:
                    type = Types(evt.header.event_type)
                except ValueError:
                    logging.warning(f"Unsupported event type {evt.header.event_type}")
                    continue

                # Idle sleep mode is encoded as an event with payload 8 when entering
                # and 09 when leaving.
                if type == Types.Event and evt.payload == b"\x08":
                    if idle_sleep_mode_started is not None:
                        logging.warning(
                            f"Idle sleep mode was already active at"
                            f" {idle_sleep_mode_started}"
                        )
                    idle_sleep_mode_started = evt.header.timestamp
                    continue
                if type == Types.Event and evt.payload == b"\x09":
                    if idle_sleep_mode_started is not None and last_values is not None:
                        payload = self._fill_ism(
                            idle_sleep_mode_started, evt.header.timestamp, last_values
                        )
                        idle_sleep_mode_started = None
                        acceleration.append(payload)
                        continue
                    else:
                        logging.warning(
                            f"Idle sleep mode was not active at {evt.header.timestamp}"
                        )
                        continue

                # An 'Activity' (id: 0x00) log record type with a 1-byte payload is
                # captured on a USB connection event (and does not represent a reading
                # from the activity monitor's accelerometer). This event is captured
                # upon docking the activity monitor (via USB) to a PC or CentrePoint
                # Data Hub (CDH) device. Therefore, such records cannot be parsed as the
                # traditional activity log records and can be ignored.
                if type in [Types.Activity, Types.Activity3]:
                    if evt.header.payload_size == 1:
                        continue

                if type == Types.Activity3:
                    payload = read_activity3_payload(evt.payload, evt.header.timestamp)
                elif type == Types.Activity2:
                    payload = read_activity2_payload(evt.payload, evt.header.timestamp)
                elif type == Types.Activity:
                    payload = read_activity1_payload(evt.payload, evt.header.timestamp)
                elif type == Types.TemperatureRecord:
                    temperature.append(
                        read_temperature_payload(evt.payload, evt.header.timestamp)
                    )
                    continue
                else:
                    continue
                if payload.shape[0] > 0:
                    last_values = payload[-1, 1:]
                    # Without the next line, if we miss an ISM stop event, we would
                    # think we are in ISM even when receiving accelerometer data.
                    idle_sleep_mode_started = None
                acceleration.append(payload)

            if idle_sleep_mode_started is not None:
                # Idle sleep mode was started but not finished before the recording
                # ended. This means that we might be missing some records at the end of
                # the file.
                idle_sleep_mode_ended = evt.header.timestamp
                payload = self._fill_ism(
                    idle_sleep_mode_started, idle_sleep_mode_ended, last_values
                )
                acceleration.append(payload)

        else:
            payload = read_nhanse_payload(
                self.activity_file,
                self.info.start_date,
                self.info.sample_rate,
            )
            acceleration.append(payload)

        if len(acceleration) > 0:
            self.acceleration = np.concatenate(acceleration)
        if len(temperature) > 0:
            self.temperature = np.concatenate(temperature)

    def to_pandas(self):
        """Return acceleration data as pandas data frame."""
        col_names = ["Timestamp", "X", "Y", "Z"]
        if len(self.acceleration) == 0:
            self.get_data()
        df = pd.DataFrame(self.acceleration, columns=col_names)
        df.set_index("Timestamp", drop=True, inplace=True)
        return df

    def temperature_to_pandas(self):
        """Return temperature data as pandas data frame."""
        col_names = ["Timestamp", "TemperatureMCU", "TemperatureADXL"]
        if len(self.temperature) == 0:
            assert len(self.acceleration) == 0
            self.get_data()
        df = pd.DataFrame(self.temperature, columns=col_names)
        df.set_index("Timestamp", drop=True, inplace=True)
        return df


class LogReader:
    """
    Handle reading GT3X/AGDC log events.

    Parameters:
        -----------
            source: IO stream for log.bin file data
    """

    def __init__(self, source):
        """Initialise reader."""
        self.source = source

    def read_event(self):
        """Parse an event."""
        header_bytes = self.source.read(8)
        if len(header_bytes) != 8:
            return None
        header = Header(header_bytes)
        payload_bytes = self.source.read(header.payload_size)
        if len(payload_bytes) != header.payload_size:
            return None
        checksum = self.source.read(1)
        if not checksum:
            return None
        try:
            raw_event = RawEvent(header, payload_bytes, checksum)
        except ValueError:
            return None
        return raw_event



"""GT3x header structure."""
import io
import logging
import struct
from dataclasses import InitVar, dataclass

import numpy as np


@dataclass
class Header:
    """
    GT3X Header.

    Attributes:
    -----------
    separator:
        Log separator value
    timestamp:
        Unix epoch timestamp in seconds
    eventType:
        GT3X event type
    payload_size:
        Event payload size in bytes
    """

    separator: bytes
    timestamp: int
    event_types: bytes
    payload_size: int

    def __init__(self, header_bytes: bytes):
        """Unpack header."""
        (separator, event_type, timestamp, payload_size) = struct.unpack(
            "<BBLH", header_bytes
        )
        self.separator = separator
        self.timestamp = timestamp
        self.event_type = event_type
        self.payload_size = payload_size


@dataclass(frozen=True)
class RawEvent:
    """
    Gt3xRawEvent class.

    Attributes:
    -----------
    header:
        Header object
    payload:
        Log event payload as byte array
    checksum:
        Log event checksum
    """

    header: Header
    payload: bytes
    checksum: InitVar[bytes]

    def __post_init__(self, checksum: bytes):
        """Verify event's checksum."""
        new_checksum = self.header.separator ^ self.header.event_type
        timestamp = self.header.timestamp.to_bytes(4, "little")
        new_checksum = np.bitwise_xor.reduce(
            np.frombuffer(timestamp, dtype=np.uint8), initial=new_checksum
        )
        payload_size = self.header.payload_size.to_bytes(4, "little")
        new_checksum = np.bitwise_xor.reduce(
            np.frombuffer(payload_size, dtype=np.uint8), initial=new_checksum
        )
        new_checksum = np.bitwise_xor.reduce(
            np.frombuffer(self.payload, dtype=np.uint8), initial=new_checksum
        )
        new_checksum = int(~new_checksum & 0xFF)
        if new_checksum.to_bytes(1, "little") != checksum:
            logging.warning(f"Corrupted event at {self.header.timestamp}.")
            raise ValueError("Event checksum does not match.")


@dataclass
class Info:
    """Metadata class."""

    start_date: int
    device_type: str
    acceleration_max: float
    sample_rate: int
    timezone: str
    subject_name: str
    battery_voltage: float
    download_date: int
    unexpected_resets: int
    last_sample_time: int
    acceleration_min: float
    acceleration_scale: float
    serial_numer: str
    limb: str
    board_revision: str
    stop_date: int
    firmware: str

    def __init__(self, zip_file):
        """Parse info.txt and returns dictionary with key/value pairs."""
        output = dict()
        with io.TextIOWrapper(
            zip_file.open("info.txt", "r"), encoding="utf-8-sig"
        ) as f:
            for line in f.readlines():
                values = line.split(":")
                # The format of TimeZone is this: "TimeZone: -04:00:00"
                if len(values) == 2 or values[0] == "TimeZone":
                    output[values[0].strip()] = ":".join(values[1:]).strip()
        self.start_date = int(output.get("Start Date", 0))
        self.device_type = output.get("Device Type", None)
        self.acceleration_max = float(output.get("Acceleration Max", 0))
        self.sample_rate = int(output.get("Sample Rate", 0))
        self.timezone = output.get("TimeZone", None)
        self.subject_name = output.get("Subject Name", None)
        self.battery_voltage = float(
            output.get("Battery Voltage", "0").replace(",", ".")
        )
        self.download_date = int(output.get("Download Date", 0))
        self.unexpected_resets = output.get("Unexpected Resets", 0)
        self.last_sample_time = int(output.get("Last Sample Time", 0))
        self.acceleration_min = float(output.get("Acceleration Min", 0))
        self.acceleration_scale = float(output.get("Acceleration Scale", 0))
        self.serial_numer = output.get("Serial Number", None)
        self.limb = output.get("Limb", None)
        self.board_revision = output.get("Board Revision", None)
        self.stop_date = int(output.get("Stop Date", 0))
        self.firmware = output.get("Firmware", None)
